/*
 * $Header: /home/orchestra5/davy/stuff/misc/xsat/RCS/satellites.c,v 1.1 92/04/10 14:08:14 davy Exp $
 *
 * Copyright 1992 by David A. Curry
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * author makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 *
 * Routines for dealing with the "satellites" menu, reading the satellite
 * information, etc.
 *
 * David A. Curry
 * Purdue University
 * Engineering Computer Network
 * 1285 Electrical Engineering Building
 * West Lafayette, IN 47907
 * davy@ecn.purdue.edu
 *
 * $Log:	satellites.c,v $
 * Revision 1.1  92/04/10  14:08:14  davy
 * Initial revision
 * 
 */
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Dialog.h>
#include <stdio.h>

#include "xsat.h"

static SatInfo	*satInfo;

static void	setSat();
static int	satCompare();

/*
 * Prompt the user for a satellite name.
 */
void
satellitesSelectSatellite(w, clientData, callData)
XtPointer clientData, callData;
Widget w;
{
	char buf[BUFSIZ];

	bzero(buf, sizeof(buf));

	if (currentSat != NULL)
		sprintf(buf, "%s", currentSat->s_name);

	prompt("Enter satellite name:", buf, setSat);
}

/*
 * Show current satellite information.
 */
void
satellitesShowSatellite(w, clientData, callData)
XtPointer clientData, callData;
Widget w;
{
	Arg argList[2];
	char buf[BUFSIZ];

	if (currentSat == NULL) {
		prompt("No satellite selected.", NULL, promptOkay);
		return;
	}

	rewind(tmpFP);
	ftruncate(fileno(tmpFP), 0);

	fprintf(tmpFP, "Satellite:       %s\n", currentSat->s_name);
	fprintf(tmpFP, "Catalog number:  %lu\n", currentSat->s_satnum);
	fprintf(tmpFP, "Epoch time:      %lf\n", currentSat->s_epochday);
	fprintf(tmpFP, "Element set:     %lu\n", currentSat->s_elementset);
	fprintf(tmpFP, "Inclination:     %lf deg\n",
		currentSat->s_inclination);
	fprintf(tmpFP, "RA of node:      %lf deg\n", currentSat->s_raan);
	fprintf(tmpFP, "Eccentricity:    %lf\n", currentSat->s_eccentricity);
	fprintf(tmpFP, "Arg of perigee:  %lf deg\n", currentSat->s_argperigee);
	fprintf(tmpFP, "Mean anomaly:    %lf deg\n",
		currentSat->s_meananomaly);
	fprintf(tmpFP, "Mean motion:     %lf rev/day\n",
		currentSat->s_meanmotion);
	fprintf(tmpFP, "Decay rate:      %le rev/day^2\n",
		currentSat->s_decayrate);
	fprintf(tmpFP, "Epoch rev:       %lu\n", currentSat->s_epochrev);

	fflush(tmpFP);

	XtSetArg(argList[0], XtNstring, tmpFile);
	XtSetValues(orbitText, argList, ONE);

	sprintf(buf, "Information for Current Satellite");
	XtSetArg(argList[0], XtNlabel, buf);
	XtSetValues(botLabel, argList, ONE);
}

/*
 * List all satellites.
 */
void
satellitesListSatellites(w, clientData, callData)
XtPointer clientData, callData;
Widget w;
{
	int i, n;
	Arg argList[2];
	char buf[BUFSIZ];
	SatInfo *si, **sip;

	rewind(tmpFP);
	ftruncate(fileno(tmpFP), 0);

	/*
	 * Sort them into alphabetical order first.
	 */
	n = 0;
	for (si = satInfo; si != NULL; si = si->s_next)
		n++;

	sip = (SatInfo **) safeAlloc(n * sizeof(SatInfo *));

	n = 0;
	for (si = satInfo; si != NULL; si = si->s_next)
		sip[n++] = si;

	qsort(sip, n, sizeof(SatInfo *), satCompare);

	for (i=0; i < n; i++)
		fprintf(tmpFP, "%-15s%d\n", sip[i]->s_name, sip[i]->s_satnum);

	fflush(tmpFP);

	XtSetArg(argList[0], XtNstring, tmpFile);
	XtSetValues(orbitText, argList, ONE);

	sprintf(buf, "List of Satellites and Catalog Numbers");
	XtSetArg(argList[0], XtNlabel, buf);
	XtSetValues(botLabel, argList, ONE);
}

/*
 * Read in the orbital elements.
 */
void
readSatList()
{
	FILE *fp;
	int linenum;
	SatInfo *si;
	char line[BUFSIZ];

	sprintf(line, "%s/%s", XSATLIBDIR, SATFILE);

	if ((fp = fopen(line, "r")) == NULL) {
		perror(line);
		exit(1);
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		line[strlen(line)-1] = '\0';

		if (satInfo == NULL) {
			satInfo = (SatInfo *) safeAlloc(sizeof(SatInfo));
			si = satInfo;
		}
		else {
			si->s_next = (SatInfo *) safeAlloc(sizeof(SatInfo));
			si = si->s_next;
		}

		stripSpace(line);
		si->s_name = saveString(line);

		fgets(line, sizeof(line), fp);
		sscanf(line, "%1d", &linenum);

		if (linenum != 1) {
			fprintf(stderr, "Line 1 missing for satellite %s.\n",
				si->s_name);
			exit(1);
		}

		sscanf(line, "%*2c%5d%*10c%2lf%12lf%10lf%*21c%5lu",
		       &si->s_satnum, &si->s_epochyear, &si->s_epochday,
		       &si->s_decayrate, &si->s_elementset);
		si->s_epochday += si->s_epochyear * 1000;
		si->s_elementset /= 10;

		fgets(line, sizeof(line), fp);
		sscanf(line, "%1d", &linenum);

		if (linenum != 2) {
			fprintf(stderr, "Line 2 missing for satellite %s.\n",
				si->s_name);
			exit(1);
		}

		sscanf(line, "%*8c%8lf%8lf%7lf%8lf%8lf%11lf%5lu",
		       &si->s_inclination, &si->s_raan, &si->s_eccentricity,
		       &si->s_argperigee, &si->s_meananomaly,
		       &si->s_meanmotion, &si->s_epochrev);
		si->s_eccentricity *= 1E-7;

		si->s_next = NULL;
	}

	fclose(fp);
}

/*
 * Find the named satellite in the SatInfo structures.
 */
SatInfo *
findSat(sname)
char *sname;
{
	SatInfo *si;

	for (si = satInfo; si != NULL; si = si->s_next) {
		if (strcmp(si->s_name, sname) == 0)
			return(si);
	}

	return(NULL);
}

/*
 * Callback for setting the satellite.
 */
static void
setSat(w, clientData, callData)
XtPointer clientData, callData;
Widget w;
{
	char *new;
	SatInfo *si;
	char buf[BUFSIZ];

	new = XawDialogGetValueString(XtParent(w));

	if ((si = findSat(new)) == NULL) {
		sprintf(buf, "Unknown satellite: %s", new);
		prompt(buf, NULL, promptOkay);
	}
	else {
		currentSat = si;
	}

	XtDestroyWidget(clientData);
}

/*
 * Comparison for qsort.
 */
static int
satCompare(a, b)
SatInfo **a, **b;
{
	return(strcmp((*a)->s_name, (*b)->s_name));
}
